/**
 * This file is part of
 * 
 * Repast Simphony MySQL Database Outputter
 *
 * Copyright (C) 2012 Argonne National Laboratory
 * Copyright (C) 2013 Center for Environmental Systems Research, Kassel, Germany
 * 
 * Repast Simphony MySQL Database Outputter is free software; you can redistribute 
 * it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; If not, see <http://www.gnu.org/licenses/>.
 * 
 * Center for Environmental Systems Research, Kassel
 * Created by Sascha Holzhauer on 13.04.2010
 */
package repast.simphony.data.db.outputter;



import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.math.BigDecimal;
import java.sql.Connection;
import java.sql.Date;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.sql.Time;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPasswordField;
import javax.swing.SwingUtilities;

import org.apache.commons.lang.StringEscapeUtils;
import org.apache.log4j.jdbc.JDBCAppender;

import repast.simphony.data.db.outputter.engine.DbOutputterDescriptor;
import repast.simphony.data2.DataSink;
import repast.simphony.engine.environment.RunEnvironment;
import repast.simphony.freezedry.FieldUtilities;
import repast.simphony.freezedry.datasource.DBConnectionInfo;
import repast.simphony.freezedry.datasource.DBUtils;
import simphony.util.messages.MessageCenter;


/**
 * 
 * repast.simphony.data.db
 * 
 * The DB outputter that delegates the output to a {@link CustomJdbcAppender}.
 * 
 * TODO check why close resp. runCleanup is not invoked (to flush) - is invoked not before run is reset
 * 
 * TODO whether check is really required - seems that close() is performed twice ; however, second call is to late... !?
 * and yields exception: "No operations allowed after connection closed."
 * 
 * @author Sascha Holzhauer
 * @date 13.04.2010
 * 
 */
public class DefaultDBOutputter implements DataSink {

	private static final MessageCenter	LOG					= MessageCenter.getMessageCenter(DefaultDBOutputter.class);

	private static final String PSEUDO_DESCRIPTOR = "Pseudo DB Outputter Descriptor";

	
		
		protected Map<String, Object> row = new HashMap<String, Object>();

		private final DbOutputterDescriptor					descriptor;

		private Connection								con;

		private DBConnectionInfo						dbConInfo;

		private int										cached				= 0;

		private PreparedStatement						preSt;

		private boolean									wroteRunInformation	= false;

		/**
		 * Stores the last runId for each output descriptor.
		 */
		private static int								runId = Integer.MIN_VALUE;
		private static Map<DefaultDBOutputter, Boolean> runIdCalculations = 
				new HashMap<DefaultDBOutputter, Boolean>();
		
		/**
		 * Initialise the DefaultDBOutputter by assigning the outputter descriptor. Uses {@link JDBCAppender} as
		 * {@link org.apache.log4j.Appender}.
		 * 
		 * @param descriptor the {@link DbOutputterDescriptor} of the current output configuration
		 */
		public DefaultDBOutputter(DbOutputterDescriptor descriptor) {
			this.descriptor = descriptor; 
			
			// this.setName(descriptor.getName() + "_" + RunState.getInstance().getRunInfo().getRunNumber());
			LOG.info("DefaultDBOutputter: Initialized with descriptor " + descriptor.getName());

			// <- LOGGING
			LOG.debug("Init CustomJdbcAppender for outpt descriptor: " + descriptor.getName(), this);
			LOG.debug("Descriptor Data Set id: " + descriptor.getDataSet(), this);
			LOG.debug("Descriptor Data Sets: " + descriptor.getDataSet(), this);
			// LOGGING ->

			
		}

		/**
		 * Return a unique run identifier that is based on the database. It retrieves a stored ID when the current
		 * descriptor was used before. In that case, the runID does not need to be incremented since it is the same run
		 * (the descriptor changes between single batch runs). Otherwise, the id is collected from the database. If there
		 * is a Repast Simphony parameter called RUN_ID, this one is incremented as long as there is already such an id in
		 * the run information table. When run information table is defined and contains rows the highest run Id is used. Otherwise, the
		 * max run ID from the data table is used. 
		 * 
		 * @param descriptor the {@link DbOutputterDescriptor} of the current output configuration
		 * @param con SQL Connection object
		 * 
		 * @return runID
		 */
		public static int getRunIDStatic(DbOutputterDescriptor descriptor, Connection con) {

			LOG.debug("Request runID: " + descriptor.getName() + "(" + descriptor.hashCode() + " / " + runId + ")");
			
			int run = runId;

			// <- LOGGING
			LOG.debug("run: " + run);
			// LOGGING ->

			if (run == Integer.MIN_VALUE) {
			
				// every time a new run is started this area is performed since the appender is renewed for every run
				// and its constructor throws Integer.MIN_VALUE into the map.
				try {
					int runIdTmp = Integer.MIN_VALUE;

					Statement statement;
					String sql;
					ResultSet result;

					boolean invalid = true;
					
					// check if RS parameter is set and assign to runIdTmp:
					if (RunEnvironment.getInstance().getParameters().getSchema().contains("RUN_ID")) {
						runIdTmp = ((Integer) RunEnvironment.getInstance().getParameters().getValue("RUN_ID"))
								.intValue();
						
						LOG.debug("RS contains RUN_ID!");
						
						if (descriptor.getRunInfoTableName() != null) {
							while (invalid) {
								// check if runIdTmp conflicts with existing runID in table:
								sql = "SELECT COUNT(*) AS 'numRows' FROM " + descriptor.getRunInfoTableName()
										+ " Where runID = " + runIdTmp;
								LOG.debug("SQL: " + sql);
								
								try {
									statement = con.createStatement();
									result = statement.executeQuery(sql);
									result.next();
									invalid = result.getInt(1) > 0;
								
									if (invalid) {
										// conflicting runID exists in table:
										LOG.warn("Repast Simphony RUN_ID parameter is already in table!", descriptor);
										runIdTmp++;
									}
									statement.close();
								} catch (SQLException e) {
									LOG.warn("RunID could not be obtaiend from table " + descriptor.getRunInfoTableName()+ "!", e);
									e.printStackTrace();
								}
							}
						}
						run = runIdTmp;						
					} 
						
					if (runIdTmp == Integer.MIN_VALUE && descriptor.getRunInfoTableName() != null) {
						// get max runID from run information table:
						sql = "SELECT MAX(runID) AS 'numRows' FROM " + descriptor.getRunInfoTableName();
						LOG.debug("SQL: " + sql);
						statement = con.createStatement();
						result = statement.executeQuery(sql);
						result.next();
						runIdTmp = result.getInt(1) + 1;
						run = runIdTmp;
					}
					
					if (runIdTmp == Integer.MIN_VALUE) {
						// check if data table contains any row:
						sql = "SELECT COUNT(*) AS 'numRows' FROM " + descriptor.getTableName();
						LOG.debug("SQL: " + sql);
						statement = con.createStatement();
						result = statement.executeQuery(sql);
						result.next();
						if (result.getInt(1) > 0) {
							// fetch maximum runId from table:
							sql = "SELECT MAX(runID) FROM " + descriptor.getTableName();
							LOG.debug("SQL: " + sql);
							statement = con.createStatement();
							result = statement.executeQuery(sql);
							result.next();
							LOG.info("Incrementing maximum runID in table...", descriptor);
							run = result.getInt(1) + 1;
						} else {
							run = 1;
						}
						statement.close();
					}

				} catch (SQLException ex) {
					ex.printStackTrace();
					LOG.warn("SQL error while fetching runID. Try using 1", ex);
					run = 1;
				}
				LOG.debug("Set runID to " + run, descriptor);
				runId = run;
			}
			return run;
		}

		/**
		 * @param descriptor
		 * @param con
		 * @return
		 */
		public int getRunID(DbOutputterDescriptor descriptor, Connection con) {
			int run = getRunIDStatic(descriptor, con);
			runIdCalculations.put(this, Boolean.TRUE);
			return run;
		}
		
		/**
		 * version information must be set as model name. paramID must be set as PARAMS_ID at RunEnvironment.
		 * 
		 * TODO check if read by another outputter!
		 * 
		 * @param run
		 * @param con
		 */
		private void writeRunInformation(int run, Connection con) {
			if (!wroteRunInformation) {
				LOG.debug("Write run information");
				// make sure that table exists:
				List<String> columns = new ArrayList<String>();
				columns.add("version");
				columns.add("timestamp");
				columns.add("paramSetID");
				try {
					if (this.descriptor.getRunInfoTableName() == null) {
						LOG.warn("No tablename for run information given in DB outputter desciptor (" +
								this.descriptor.getName() + ")", this);
						throw new IllegalStateException("No tablename for run information given in DB outputter desciptor (" +
								this.descriptor.getName() + ")");
					}
					createTableIfRequired(this.descriptor.getRunInfoTableName(), columns);

					String versionInfo;
					if (!RunEnvironment.getInstance().getParameters().getSchema().contains("VERSION")) {
						LOG.warn("RS Parameter 'VERSION' is no set. Using default value", this);
						versionInfo = "No version information available";
					} else {
						versionInfo = (String) RunEnvironment.getInstance().getParameters().getValue("VERSION");
					}
					
					String paramSetID;
					if (!RunEnvironment.getInstance().getParameters().getSchema().contains("PARAM_SET_ID")) {
						LOG.warn("RS Parameter 'PARAM_SET_ID' is no set. Using 0", this);
						paramSetID = "0";
					} else {
						paramSetID = ((Integer) RunEnvironment.getInstance().getParameters().getValue("PARAM_SET_ID")).toString();
					}

					// check if another outputter already send run information
					String sql = "SELECT COUNT(*) AS 'numRows' FROM " + descriptor.getRunInfoTableName()
					+ " Where runID = " + run;
					LOG.debug("SQL: " + sql);
					Statement statement = con.createStatement();
					ResultSet result = statement.executeQuery(sql);
					result.next();

					if (result.getInt(1) == 0) {
						sql = "INSERT INTO " + descriptor.getRunInfoTableName() + " SET " + "runID = " + run + ", "
								+ "version = '" + versionInfo + "', "
								+ "timestamp = NOW(), " + "paramSetID = "
								+ paramSetID + ";";
						LOG.debug("SQL: " + sql);
						statement = con.createStatement();
						statement.executeUpdate(sql);
						statement.close();
					}
					else {
						LOG.debug("There already exists an entry for run " + run + " in " + descriptor.getRunInfoTableName());
					}

				} catch (SQLException e) {
					LOG.warn("Creation of run information failed", e);
					e.printStackTrace();
				}
				wroteRunInformation = true;
			}
		}

		/**
		 * @see org.apache.log4j.jdbc.JDBCAppender#getConnection()
		 * 
		 * A connection pool (using Apache DBCP, for instance) seems to be overload,
		 * since only one connection is used throughout the simulation run.
		 */
		public Connection getConnection() {
			if (this.con == null) {
				this.con = getConnection(null, dbConInfo);
			}
			return this.con;
		}
		
		public static Connection getConnection(Connection con, DBConnectionInfo dbConInfo) {
			if (con == null) {

				try {
					Class.forName("com.mysql.jdbc.Driver").newInstance();

					con = DriverManager.getConnection(dbConInfo.dbURL, dbConInfo.userName, dbConInfo.password);

					LOG.debug("Connection " + con + " established.", dbConInfo);

				} catch (SQLException ex) {
					LOG.error("CustomJdbcAppender: SQL error while requesting connection.", ex);
				} catch (InstantiationException e) {
					LOG.error("CustomJdbcAppender: Error while requesting connection.", e);
					e.printStackTrace();
				} catch (IllegalAccessException e) {
					LOG.error("CustomJdbcAppender: Error while requesting connection.", e);
					e.printStackTrace();
				} catch (ClassNotFoundException e) {
					LOG.error("CustomJdbcAppender: Error while requesting connection.", e);
					e.printStackTrace();
				}
			}
			LOG.debug("CustomJdbcAppender: Connection returned: " + con);
			return con;
		}

		/**
		 * @return Created by Sascha Holzhauer on 18.03.2011
		 */
		protected PreparedStatement getPreparedStatement() {
			if (preSt == null) {
				StringBuffer sql = new StringBuffer();
				StringBuffer sqlValues = new StringBuffer();
				sql.append("INSERT INTO ");
				sql.append(getCleanName(descriptor.getTableName()));
				sql.append(" (");

				sqlValues.append("VALUES(");

				// append runID column to header:
				if (descriptor.isAddRunID()) {
					sql.append("RunID");
					sql.append(", ");
					sqlValues.append("?,");
				}

				// append columns as defined in output configuration to header:
				Iterator<String> columns = descriptor.getSourceIds().iterator();
				while (columns.hasNext()) {
					String column = columns.next();
					sql.append(getCleanName(column));
					if (columns.hasNext()) {
						sql.append(", ");
						sqlValues.append("?,");
					} else {
						sqlValues.append("?");
					}
				}
				sql.append(") ");
				sql.append(sqlValues);
				sql.append(")");

				// <- LOGGING
				LOG.debug("Prepare Statement: " + sql.toString());
				// LOGGING ->

				try {
					preSt = getConnection().prepareStatement(sql.toString());
				} catch (SQLException e) {
					LOG.warn("Creation of SQL prepare statement failed.", e);
					e.printStackTrace();
				}
			}
			return preSt;
		}


		/**
		 * Executes the customBuffer directly to super.execute() to ensure customBuffer is executed no matter if cache
		 * size is reached.
		 * 
		 * 
		 * @see org.apache.log4j.jdbc.JDBCAppender#close()
		 */
		@Override
		public void close() {
			LOG.info(descriptor.getName() + "(" + descriptor.hashCode() + "): Appender closing");
	
			try {
				if (!con.isClosed()) {
					LOG.debug("Execute Batches");
					PreparedStatement prepStat = getPreparedStatement();
					prepStat.executeBatch();
	
					if (prepStat != null) {
						prepStat.close();
					}
				}
			} catch (SQLException ex) {
				ex.printStackTrace();
				LOG.error("CustomJdbcAppender: SQL error while executing for closing.", ex);
			}

			try {
				if (con != null && !con.isClosed()) {
					LOG.debug("Connection " + con + " closing.", this);
					con.close();
				}

			} catch (SQLException e) {
				LOG.error("Error closing connection", e);
			}
		}

		/**
		 * @see org.apache.log4j.jdbc.JDBCAppender#finalize()
		 */
		@Override
		public void finalize() {
			// prevent super class from calling close();
		}

		@Override
		public void append(String key, Object value) {
			this.row.put(key, value);
		}
		
		/**
		 * @see org.apache.log4j.jdbc.JDBCAppender#getLogStatement(org.apache.log4j.spi.LoggingEvent)
		 */
		public String getLogStatement() {

			if (!wroteRunInformation) {
				writeRunInformation(getRunID(descriptor, getConnection()), getConnection());
			}

			PreparedStatement prepStat = getPreparedStatement();
			int index = 0;

			try {
				if (descriptor.isAddRunID()) {
					prepStat.setInt(++index, getRunID(descriptor, getConnection()));
				}

				Iterator<String> columns = descriptor.getSourceIds().iterator();
				while (columns.hasNext()) {
					String column = columns.next();
					addToPrepStatet(prepStat, ++index, row.get(column));
				}
				prepStat.addBatch();

				if (descriptor.getWriteCacheInterval() == cached) {
					// the number of records to cache is reached -> execute statement
					LOG.debug("Execute buffered batches");
					prepStat.executeBatch();
					cached = 0;
				} else {
					cached++;
					LOG.debug("Cached incremented: " + cached);
				}
			} catch (SQLException ex) {
				LOG.warn("Execution of SQL failed.", ex);
			}

			// this method must override a super type method...
			return "";
		}

		protected String getValue(String value) {
			return "\'" + StringEscapeUtils.escapeJava(value) + "\'";
		}

		protected String getValue(char[] value) {
			return getValue(new String(value));
		}

		protected String getValue(Number number) {
			if (number.equals(Double.NaN) || number.equals(Float.NaN)) {
				return "Null";
			}
			return number.toString();
		}

		protected String getValue(Boolean value) {
			return value.toString();
		}

		protected String getCleanName(String type) {
			return type.replaceAll("\\.|\\[|\\(", "_").replaceAll("\"", "");
		}

		/**
		 * @param value
		 * @return 
		 */
		public String get(Object value) {
			if (value instanceof String) {
				return getValue((String) value);
			} else if (value instanceof char[]) {
				return getValue((char[]) value);
			} else if (value instanceof Number) {
				return getValue((Number) value);
			} else if (value instanceof Boolean) {
				return getValue((Boolean) value);
			} else {
				LOG.debug("Other value type to convert: " + value + "(" + value.getClass() + ")");
				return "NULL";
			}
		}

		/**
		 * @param prepStat
		 * @param index
		 * @param value
		 *
		 * @throws SQLException
		 */
		public void addToPrepStatet(PreparedStatement prepStat, int index, Object value) throws SQLException {
			
			// <- LOGGING
				LOG.debug("Add to PrepStatment: " + value + "(" + index + ")");
			// LOGGING ->

			if (value instanceof Double) {
				if (((Double) value).isNaN() || ((Double) value).isInfinite()) {
					prepStat.setNull(index, java.sql.Types.DOUBLE);
				} else {
					prepStat.setDouble(index, (Double) value);
				}
			} else if (value instanceof Float) {
				if (((Float) value).isNaN() || ((Float) value).isInfinite()) {
					prepStat.setNull(index, java.sql.Types.FLOAT);
				} else {
					prepStat.setFloat(index, (Float) value);
				}
			} else if (value instanceof Integer) {
				prepStat.setInt(index, (Integer) value);
			} else if (value instanceof Long) {
				prepStat.setLong(index, (Long) value);
			} else if (value instanceof String) {
				prepStat.setString(index, (String) value);
			} else if (value instanceof Short) {
				prepStat.setShort(index, (Short) value);
			} else if (value instanceof Boolean) {
				prepStat.setBoolean(index, (Boolean) value);
			} else if (value instanceof BigDecimal) {
				prepStat.setBigDecimal(index, (BigDecimal) value);
			} else if (value instanceof Date) {
				prepStat.setDate(index, (Date) value);
			} else if (value instanceof Time) {
				prepStat.setTime(index, (Time) value);
			} else if (value instanceof char[]) {
				prepStat.setString(index, ((char[]) value).toString());
			} else {
				LOG.warn("Other value type to convert: " + value + "(" + value.getClass() + ")");
			}
		}

		/**
		 * Checks whether the given table exists (does not check for correct columns!). If not, it creates a new table
		 * with the given column names of type "VARCHAR(1024)".
		 * 
		 * @param tableName
		 * @param colNames
		 * @throws SQLException
		 */
		protected void createTableIfRequired(String tableName, List<String> cols) throws SQLException {

			List<String> colNames = new ArrayList<String>();
			for (String s : cols) {
				colNames.add(s);
			}

			if (!DBUtils.doesTableExist(getConnection(), getCleanName(tableName))) {

				StringBuilder builder = new StringBuilder("Create table ");

				builder.append(getCleanName(tableName));
				builder.append(" (");

				if (descriptor.isAddRunID()) {
					builder.append("runID INT(10)");
					builder.append(", ");
				}

				for (int i = 0; i < colNames.size(); i++) {
					builder.append(getCleanName(colNames.get(i)));
					builder.append(" ");
					if (getCleanName(colNames.get(i)).equals("timestamp")) {
						builder.append("TIMESTAMP");
					} else {
						builder.append("VARCHAR(1024)");
					}
					if (i < colNames.size() - 1) {
						builder.append(", ");
					}
				}
				builder.append(")");

				LOG.info("Creating table " + tableName + ")");
				executeCreate(getConnection(), builder.toString());
			} else {
				// TODO doesTableDefinitionExist(String tableName, Map<String, Object> cols)
				if (!doesTableDefinitionExist(tableName, colNames)) {
					StringBuffer buffer = new StringBuffer();
					for (String col : colNames) {
						buffer.append("\t" + col + "\n");
					}
					LOG.warn("Table definition (" + tableName + ") does not meet requirements: \n" + buffer.toString(),
							this);
				}
			}
		}

		/**
		 * Checks whether a table with the given name exists. If yes, it checks whether
		 * the table contains the columns of the specified name.
		 * 
		 * @param tableName the name of the table to check
		 * @param list of String representing the columns to check for
		 * 
		 * @return true if a table exists that serves the given columns
		 * 
		 * @throws SQLException
		 */
		protected boolean doesTableDefinitionExist(String tableName, List<String> cols) throws SQLException {

			// does a table with given name exist?
			if (!DBUtils.doesTableExist(getConnection(), getCleanName(tableName))) {
				return false;
			}

			if (cols.size() == 0 && tableName.startsWith("\"") && tableName.endsWith("\"") && tableName.length() > 1) {
				tableName = tableName.substring(1, tableName.length() - 1);
			}

			// check if every column in cols exists in table:
			ResultSet set = getConnection().getMetaData().getColumns(null, null, tableName, "%");

			Map<String, Integer> tableCols = new HashMap<String, Integer>();
			while (set.next()) {
				tableCols.put(set.getString("COLUMN_NAME").toUpperCase(), new Integer(set.getInt("DATA_TYPE")));
			}

			for (String entry : cols) {

				if (!tableCols.containsKey(entry.toUpperCase())) {
					return false;
				}
			}
			return true;
		}

		/**
		 * Checks whether a table with the given name exists. If yes, it checks whether
		 * the table contains the columns of the specified name and of the specified type.
		 * 
		 * @param tableName the name of the table to check
		 * @param cols map of String/Object entries representing the columns to check for
		 * 
		 * @return true if a table exists that serves the given columns
		 * 
		 * @throws SQLException
		 */
		protected boolean doesTableDefinitionExist(String tableName, Map<String, Object> cols) throws SQLException {

			// does a table with given name exist?
			if (!DBUtils.doesTableExist(getConnection(), getCleanName(tableName))) {
				return false;
			}

			if (cols.size() == 0 && tableName.startsWith("\"") && tableName.endsWith("\"") && tableName.length() > 1) {
				tableName = tableName.substring(1, tableName.length() - 1);
			}

			// check if every column in cols exists in table:
			ResultSet set = getConnection().getMetaData().getColumns(null, null, tableName, "%");

			Map<String, Integer> tableCols = new HashMap<String, Integer>();
			// TODO make case insensitive!
			while (set.next()) {
				tableCols.put(set.getString("COLUMN_NAME"), new Integer(set.getInt("DATA_TYPE")));
			}

			for (Entry<String, Object> entry : cols.entrySet()) {

				if (!tableCols.containsKey(entry.getKey())
						|| (getColumnTypeInt(entry.getValue()) != tableCols.get(entry.getKey()))) {
					return false;
				}
			}
			return true;
		}

		/**
		 * Identifies the SQL type statement for the type of the given value.
		 * 
		 * @param value
		 * @return SQL type statement
		 */
		protected int getColumnTypeInt(Object value) {
			if (value instanceof String) {
				return java.sql.Types.VARCHAR;
			} else if (value != null && FieldUtilities.INSTANCE.isPrimitiveArray(value.getClass())) {
				return java.sql.Types.ARRAY;
			} else if (value instanceof Double) {
				return java.sql.Types.DOUBLE;
			} else if (value instanceof Float) {
				return java.sql.Types.FLOAT;
			} else if (value instanceof Number) {
				return java.sql.Types.NUMERIC;
			} else if (value instanceof byte[]) {
				return java.sql.Types.BLOB;
			} else if (value instanceof Boolean) {
				return java.sql.Types.BOOLEAN;
			}
			LOG.warn("Couldn't determine type for object '" + value + "' of type "
					+ (value != null ? value.getClass() : null) + ".");
			return 0;
		}

		/**
		 * Identifies the SQL type statement for the type of the given value.
		 * 
		 * @param value
		 * @return SQL type statement
		 */
		protected String getColumnType(Object value) {
			if (value instanceof String) {
				return "VARCHAR(1024)";
			} else if (value != null && FieldUtilities.INSTANCE.isPrimitiveArray(value.getClass())) {
				return "VARCHAR(1024)";
			} else if (value instanceof Double) {
				return "DOUBLE";
			} else if (value instanceof Float) {
				return "FLOAT";
			} else if (value instanceof Number) {
				return "NUMERIC";
			} else if (value instanceof byte[]) {
				return "BLOB";
			} else if (value instanceof Boolean) {
				return "BOOLEAN";
			}
			LOG.warn("Couldn't determine type for object '" + value + "' of type "
					+ (value != null ? value.getClass() : null) + ".");
			return null;
		}

		/**
		 * Executes the given SQL-statement on connection con to create a new table and closes the connection.
		 * 
		 * @param con
		 * @param creationSQL
		 * @throws SQLException
		 */
		protected void executeCreate(Connection con, String creationSQL) throws SQLException {
			Statement statement = null;
			try {
				LOG.info("Creating table with string: '" + creationSQL + "'.");

				statement = con.createStatement();
				statement.execute(creationSQL);
				statement.close();
			} catch (SQLException ex) {
				throw ex;
			} finally {
				if (statement != null) {
					statement.close();
				}
			}
		}

		/**
		 * Checks if the current dbConInfo defines a valid SQL connection
		 * 
		 * @return true if dbConInfo defines a valid SQL connection
		 */
		protected boolean validateConnectionSettings() {
			return DBUtils.validateConnectionSettings(dbConInfo, "");
		}

		/**
		 * @return maximum runID over all output descriptors.
		 */
		public static int getCurrentMaxRunId() {
			for (DefaultDBOutputter a : runIdCalculations.keySet()) {
				LOG.debug("Check runID of " + a, a);
				if (!runIdCalculations.get(a)) {
					// make sure that current run ID is investigated:
					a.getRunID(a.descriptor, a.getConnection());
				}
			}
			return runId;
		}
		
		/**
		 * @return maximum runID over all output descriptors.
		 */
		public static int getCurrentMaxRunId(String runInfoTable, DBConnectionInfo dbConInfo) {
			if (runIdCalculations.size() == 0) {
				DbOutputterDescriptor pseudoDescriptor = new DbOutputterDescriptor();
				pseudoDescriptor.setName(PSEUDO_DESCRIPTOR);
				pseudoDescriptor.setRunInfoTableName(runInfoTable);
				Connection con = getConnection(null, dbConInfo);
				
				LOG.debug("Connection in getCurrentMaxRunID: " + con, con);
				runId = getRunIDStatic(pseudoDescriptor, con);
				try {
					con.close();
				} catch (SQLException e) {
					LOG.error("Error during close() of connection. " + con, e);
					e.printStackTrace();
				}
				
			} else {
				for (DefaultDBOutputter a : runIdCalculations.keySet()) {
					LOG.debug("Check runID of " + a, a);
					if (!runIdCalculations.get(a)) {
						// make sure that current run ID is investigated:
						a.getRunID(a.descriptor, a.getConnection());
					}
				}
			}
			return runId;
		}
	

	@Override
	public void open(List<String> sourceIds) {
		String username = "";
		String password = "";

		// prompt for user name if not set:
		if (descriptor.getUsername() == null || descriptor.getUsername().equals("")) {
			username = JOptionPane.showInputDialog("Please enter the username for database "
					+ descriptor.getDbURL() + ".");
		} else {
			username = descriptor.getUsername();
		}

		// prompt for password if not set:
		if (descriptor.getPassword() == null || descriptor.getPassword().equals("")) {
			JLabel label = new JLabel("Please enter your password for username " + username + " (database "
					+ descriptor.getDbURL() + ").");
			final JPasswordField jpf = new JPasswordField();
			JOptionPane jop = new JOptionPane(new Object[] { label, jpf }, JOptionPane.QUESTION_MESSAGE,
					JOptionPane.OK_CANCEL_OPTION);
			JDialog dialog = jop.createDialog("Password...");
			dialog.addComponentListener(new ComponentAdapter() {
				@Override
				public void componentShown(ComponentEvent e) {
					SwingUtilities.invokeLater(new Runnable() {
						@Override
						public void run() {
							jpf.requestFocusInWindow();
						}
					});
				}
			});
			dialog.setVisible(true);
			int result = (Integer) jop.getValue();
			dialog.dispose();
			if (result == JOptionPane.OK_OPTION) {
				password = new String(jpf.getPassword());
			}
		} else {
			password = descriptor.getPassword();
		}

		// check database connection:
		this.dbConInfo = new DBConnectionInfo(username, password, descriptor.getDbURL(), descriptor.getDriverName());
		// validateConnectionSettings();

		LOG.debug("CustomJdbcAppender: dbURL: " + descriptor.getDbURL());
		LOG.debug("CustomJdbcAppender: user: " + username);
		LOG.debug("CustomJdbcAppender: password: " + password);

		// reset when a new outputter is added to force new investigation for all (other) outputters: 
		runId = Integer.MIN_VALUE;
		LOG.debug("RunID: " + runId, descriptor);

		// create the table if it does not exist:
		try {
			createTableIfRequired(this.descriptor.getTableName(), this.descriptor.getSourceIds());
		} catch (SQLException ex) {
			LOG.warn("SQL error while creating missing table.", ex);
		}
		runIdCalculations.put(this, new Boolean(false));
		LOG.info("CustomJdbcAppender: Initialized");
	}

	@Override
	public void rowStarted() {
		// do nothing
	}

	@Override
	public void rowEnded() {
		getLogStatement();
		row.clear();
	}

	/**
	 * @see repast.simphony.data2.DataSink#recordEnded()
	 */
	@Override
	public void flush() {
		try {
			if (!con.isClosed()) {
				LOG.debug("Execute buffered batches");
				PreparedStatement pstate = getPreparedStatement();
				if (pstate != null) {
					pstate.executeBatch();
				}
				cached = 0;
			} else {
				LOG.debug("Connection and/or Prepared Statement already closed!");
			}
		} catch (SQLException ex) {
			LOG.warn("Execution of SQL failed.", ex);
			ex.printStackTrace();
		}
	}

	@Override
	public void recordEnded() {
		// nothing to do
	}
}
